/* 
 * A conditional PCA simulator
 *
 * ProgramName: cpca.c
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: cpca.c,v 1.2 2001/11/18 10:18:46 asakawa Exp asakawa $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "general.h"
#include "rand_gen.h"
#include "fileio_util.h"
#include "matrix_util.h"

struct NEURON {
    double output;              /* Output values of this neuron */
    int Ninp;                   /* Number of input neurons */
    double *inp_wgt;            /* array of input weights */
    struct NEURON **inp_neuron; /* the pointer to the input neurons */
    double (*output_func)();    /* the pointer to the output function */
};
typedef struct NEURON neuron;


/* ------------------------------------------------------------------------
   initial set of global variables
   ------------------------------------------------------------------------ */
int FROM=0;
int TO=1000;
int seed=0;
int EPOCH=20;

double learning_ratio = 0.005;
double rand_range = 0.1;

/* NOTE: 0-th neuron will be handled as a threshold value
   The actual number of neurons in each layer will be automatically
   settled by program in accordance with both the command line arguments
   and input file. */
int N_of_input_layer = 0;
int N_of_output_layer = 0;

neuron *input, *output;

int NPATS = 0;               /* Number of patterns to be learned */
char *input_file_name = NULL;
char *init_file_name = NULL;
matrix example;
matinfo Input_Info;

/* linear output function */
double linear_f(double state);

/* initialize a neuron which has N inputs */
void initialize_neuron(neuron *a, int N, double (*output_func)(), double range);

void calc_output(neuron *a);

void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A conditional PCA simulator\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s -output <int> -input <filename> [-init <initial_weight_file_name>] [-learn <double>] [-epoch <int>] [-to <int>] [-seed <int>] [-rand_range <double>]\n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     output, Number of output layer\n");
    fprintf(stderr, "     input, input file name to be learned\n");
    fprintf(stderr, "default settings of options: \n");
    fprintf(stderr, "         learn = %g: learning ratio\n", learning_ratio);
    fprintf(stderr, "         epoch = %d: epoch \n", EPOCH);
    fprintf(stderr, "         to = %d: max iterations\n", TO);
    fprintf(stderr, "         rand_range: = %g: range for initial values\n", rand_range);
    fprintf(stderr, "         seed: seed of random number generator\n");
}

double linear_f(double state)
{
    return state;
}

/* Initilazing a neuron, which is described as a struct NEURON */
/* in neuron_util.h, and assigning an initial value as default */
/* to each variable  */
void initialize_neuron(neuron *a, int N, double output_func(), double range )
{
    int i;

    a->output = 0.0;
    a->Ninp = N;
    a->output_func = output_func;

    if ( a->Ninp == 0 ) {
        a->inp_wgt = NULL;
        a->inp_neuron = NULL;
        return;
    }

    a->inp_wgt = (double *)malloc(sizeof(double) * (a->Ninp));
    if ( a->inp_wgt == NULL){
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i< a->Ninp; i++) {
        a->inp_wgt[i] = 2.0 * MTgenrand_f() * range - range;
    }

    a->inp_neuron = (neuron **)malloc(sizeof(neuron *) * (a->Ninp));
    if ( a->inp_neuron == NULL ) {
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i < a->Ninp; i++) {
        a->inp_neuron[i] = NULL;
    }
}

void print_neuron(neuron *a)
{
    int i;

    fprintf(stderr, "### output :%5.3f  ", a->output);
    fprintf(stderr, "### Ninp   :%d\n", a->Ninp);
    for (i=0; i<  a->Ninp; i++) {
	fprintf(stderr, "wgt[%d]=%5.3f\t", i, a->inp_wgt[i]);
    }
    fprintf(stderr, "\n");
}


void calc_output (neuron *a) 
{
    double wrk = 0.0;
    int i;

    for(i=0; i< a->Ninp; i++){
	wrk += a->inp_wgt[i] * a->inp_neuron[i]->output;
    }
    a->output = a->output_func(wrk);
}

int ParseArgs(int argc, char **argv)
/* Parse command line arguments and set the variables */
/* The variables to be set are 
   learning_ratio, seed, 
   EPOCH, N_of_output_layer, TO, 
   input_file_name, init_file_name */
{
    int argn=1;

    while ( argn < argc ) {
        if ((strcmp(argv[argn],"--help") == 0) || (strcmp(argv[argn],"-h") == 0)) {
            usage(argv[0]);     exit (EXIT_FAILURE);
        }
        if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
            if (strcmp(argv[argn], "-learn") == 0) {
                learning_ratio = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-rand_range") == 0) {
                rand_range = atof(argv[++argn]);
            }
            if (strcmp(argv[argn],"-seed") == 0) {
                seed = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-epoch") == 0) {
                EPOCH = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-output") == 0) {
                N_of_output_layer = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-to") == 0) {
                TO = atoi(argv[++argn]);
		if ( TO < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -to parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-input") == 0) {
		input_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (input_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in main()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(input_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		input_file_name[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-init") == 0) {
		init_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (init_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in main()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(init_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in main()\n");
		    exit (EXIT_FAILURE);
		}
		init_file_name[strlen(argv[argn])] = '\0';
	    }
	}
	argn++;
    }

    if ( N_of_output_layer == 0 ) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid number of output layer(%d) !!!\n", N_of_output_layer);
	exit (EXIT_FAILURE);
    }

    if (input_file_name == NULL) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid filename of input=%s !!!\n", input_file_name);
	exit (EXIT_FAILURE);
    }
    Input_Info = matinfo_read(input_file_name);

    N_of_input_layer = Input_Info->m;
    if ( N_of_input_layer <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    if ( N_of_output_layer == 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    NPATS = Input_Info->n;
    example = Input_Info->mat;

    return OK;
}

/* Set values for an input pattern to  each input neron */
/* where pattern number will be given by second argument pat(int) */
void set_input(neuron *input, int pat) 
{
    int i;
    
    for (i=0; i< N_of_input_layer; i++) {
	input[i].output = *(example+offs(NPATS,N_of_input_layer,pat+1,i+1));
    }
}

void Initialize_network(void) 
{
    int i, j;
    FILE *fp;
    matrix TMP;

    input = (neuron *)malloc(sizeof(neuron) * (N_of_input_layer));
    if ( input == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }
    output = (neuron *)malloc(sizeof(neuron) * (N_of_output_layer));
    if ( output == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }

    /* initialize each neuron in output layer */
    for (i=0; i< N_of_output_layer; i++) { 
        initialize_neuron(&output[i], N_of_input_layer, linear_f, rand_range); 
    }
    
    /* initialize each neuron in input layer */
    /* input[0] will be treated as a threshold in the output layer */
    initialize_neuron(&input[0],0, linear_f, rand_range);   
    for (i=0; i< N_of_input_layer; i++) { 
        initialize_neuron(&input[i],0, linear_f, rand_range); 
    }
    /* NOTE: */    
    /* The output values both input[0] and output[0] are              */
    /* dealed as threshold values, therefore, the output of each      */
    /* should be constantly set to 1.                                 */
    
    /********* Making connection among neurons **********/
    /* From input layer to output layer */
    for (i=0; i< N_of_output_layer; i++) {  
        for (j=0; j< N_of_input_layer; j++) {
            output[i].inp_neuron[j] = &input[j]; 
        }
    }

    /* When init_file_name will be given, then          */
    /* read the initial value of weight value from file */
    if ( init_file_name != NULL ) {
	TMP = matinit(1,N_of_input_layer);
	fp = fileio_util_openr(init_file_name);

	for (i=0; i< N_of_output_layer; i++) {  
	    freadmat(fp, TMP, 1, N_of_input_layer);
	    for (j=0; j< N_of_input_layer; j++) {
		output[i].inp_wgt[j] = *(TMP+offs(1,N_of_input_layer,1,j+1));
	    }
        }
	free(TMP);
	fileio_util_close(fp);
    }
}

/* Compute weight values */
void update_weight(neuron *a)
{
    int i;
    double dwt;

    dwt = 0.0;
    for (i=0; i< a->Ninp; i++) {
	dwt = a->output * ( a->inp_neuron[i]->output - a->inp_wgt[i] );
	a->inp_wgt[i] +=  learning_ratio * dwt;
    }
}

int do_shuffle(int *order,int n)
{
    int i, index, tmp;

    for (i=0; i<n; i++) {
        index = i + (int)(MTgenrand_f() * ((double)n - (double)i ));
	tmp = order[index];
	order[index] = order[i];
	order[i] = tmp;
    }
    return 0;
}

/* main loop of computation */
int main_loop () 
{
    int t,i, pat;
    int *shuffle;

    shuffle=(int *)malloc(sizeof(int) * NPATS);
    for ( i=0; i<NPATS; i++ ) {
	shuffle[i] = i;
    }

    for (t=FROM; t<TO; t++){ /* learning time */
	if ( (t % EPOCH) == 0 ) {
    	    for (i=0; i< N_of_output_layer; i++) {
		print_neuron(&output[i]);
	    }
	}
	do_shuffle(shuffle,NPATS);
        for (pat=0; pat<NPATS; pat++){ /* foreach pattern */
            set_input(input, shuffle[pat]);
	    
            for (i=0; i< N_of_output_layer; i++) {
                calc_output(&output[i]);
            }

            for (i=0; i< N_of_output_layer; i++) {
                update_weight(&output[i]);
            }
        }
    }
    return t;
}

int main(int argc, char **argv) 
{
    long int iterations = 0;

    ParseArgs(argc, argv);

    init_random(seed);
    Initialize_network();

    iterations = main_loop();

    fprintf(stdout, "### %s -output %d -input %s %s %s -learn %g -epoch %d -to %d -seed %d -rand_range %g\n",
	    argv[0], N_of_output_layer, input_file_name, 
            (init_file_name != NULL) ? "-init" : "",
            (init_file_name != NULL) ? init_file_name : "",
	    learning_ratio, EPOCH, TO, get_seed(), rand_range);

    return OK;
}
