/* 
 * A Hierarchical Mixtures of Experts with two layers
 * A gradient Ascend Algorithm
 *
 * Bibliography: Michael I. Jordan and Rovert A. Jacobs, (1994), 
 * Hierarchical Mixuters of Experts and the EM Algorithm,
 * Neural Comoputation, 6, 181-214.
 *
 * ProgramName: hme2.c 
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: hme2.c,v 1.2 2002/09/23 14:03:08 asakawa Exp asakawa $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "general.h"
#include "rand_gen.h"
#include "fileio_util.h"
#include "matrix_util.h"

/*******************************************************************
 * definition of a neuron
 *******************************************************************/
typedef struct neuron {
    double output;              /* Output values of this neuron */
    double state;               /* internal state */
    double prior;               /* prior probability */
    double posterior;           /* posterior probability */
    int Ninp;                   /* Number of input neurons */
    double *inp_wgt;            /* array of input weights */
    struct neuron **inp_neuron; /* the pointer to the input neurons */
    void (*output_func)();      /* the pointer to the output function */
} neuron;

typedef struct EN { /* for expert network */
    neuron *vector;
} EN;

typedef struct GN { /* for gating network */
    neuron *member;
} GN;

neuron *output_vector, *input_vector; 

/*******************************************************************
 *  The global variables for the network architecture
 *******************************************************************/
/* Each neuron in an expert network is refered as experts[i][j].vector[k] */
EN **experts, **Uij; 

/* Gateing network */
GN G1, *G2;

/* probability matrix */
matrix Pij;

/*******************************************************************
 *  declarations of the global variables
 *******************************************************************/
FLAG REGRESSION=YES;
FLAG CLASSIFICATION=NO;

int TO=10000;         /* total number of training */
int SEED=0;           /* to keep the seed of rundom number generator */
int EPOCH=20;

double ALLOWABLE_ERROR = 0.05;
double ERROR_LIMIT = 0.0;
double LEARNING_RATIO = 0.1;
double RAND_RANGE = 0.1;
double SIGMOID_SLOPE = 4.0;
double SIGMA=1.0;

/*
 * The actual number of neurons in each layer will be automatically
 * settled by program in accordance with both the command line arguments
 * and input file. 
*/
int N_of_input = 0;
int N_of_output = 0;
int N_of_top = 2;            /* default number of the top node */
int N_of_second = 2;         /* default number of the second node */
int Npats = 0;               /* Number of patterns to be learned */

char *INPUT_FILE_NAME = NULL;
char *TEACHER_FILE_NAME = NULL;
char *INIT_FILE_NAME = NULL;
matrix Example, Teacher;
matinfo Input_Info, Teacher_Info;

/*******************************************************************
 *  declarations of functions 
 *******************************************************************/
/* show usage of this program */
void usage( char *prgname );

/* Parse command line arguments and set the global variables */
int ParseArgs( int argc, char **argv );

/* Initialzing a neuron defined as a struct neuron */
void initialize_a_neuron( neuron *a, int Ninp, void output_func() );

/* McMalloch and Pitts' formal neuron */ 
void formal_f( neuron *a );

/* Logistic function $f(x)=1/(1+ exp(-x))$ */
void logistic_f( neuron *a );

/* linear output function */
void linear_f( neuron *a );

void build_experts_network( void );
void build_gating_network( void );
void build_network( void );

/* Set values for an input pattern to each input neron, */
/* where pattern number will be given by second argument pat(int) */
void set_input( int pat );

/* print information of a neuron */
void print_a_neuron( neuron *a );

/* calc softmax function, which will be called from calc_output_vector() */
void softmax( GN *gates, int n );

/* calc probability $P_{ij}(y|x,\theta)$ */
double calc_Pij( EN a, int pat );

/* calc an output correspondings to an input vector */
void calc_an_output( neuron *a );
void calc_posterior( void );
void calc_output_vector( int pat );

/* calc Sum of Square of Errors */
double calc_SSE( void );

/* genrate a randomized sequence for stochastic learning */
void do_shuffle( int *order, int n );

/* Print weights to the stdout */
void print_weights( void );
/* Print results to the stderr */
void print_results( void );

/* Print an output correspondings to a given input pattern */
void print_an_output( int pat );

/* learning of the top gating network */
/* $v_i += \rho ( h_i - g_i ) x^t$ */
void update_G1( void );

/* learning of the second gating network */
/* $v_{ij} += \rho h_i ( h_{j|i} - g_{j|i} ) x^t$ */
void update_G2( void );

/* learning of experts network */
/* $U_{ij} += \rho h_i h_{j|i} ( y - u ) x^t$ */
void update_EN( int pat );

/* main loop of computation */
int main_loop( double *SSE );

/*****************************************************************
 * Implementation of each function 
 *****************************************************************/
void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A Hierarchical Mixtures of Experts\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s [-classification|-regression] -input <filename> -teacher <filename> [-init <initial_weight_file_name>] [-top <int>] [-second <int>] [-error <double>] [-learn <double>] [-epoch <int>] [-to <int>] [-seed <int>] [-rand_range <double>] [-slope <double>] [-allowable_error <double>] \n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     -regression (default) or -classification\n");
    fprintf(stderr, "     input, input file name to be learned\n");
    fprintf(stderr, "     teacher, teacher file name to be learned\n");
    fprintf(stderr, "default settings of options: \n");
    fprintf(stderr, "         top = %d: # of top gating network\n", N_of_top);
    fprintf(stderr, "         second = %d: # of second gating network\n", N_of_second);
    fprintf(stderr, "         learn = %g: learning ratio\n", LEARNING_RATIO);
    fprintf(stderr, "         error = %g: error limit\n", ERROR_LIMIT);
    fprintf(stderr, "         allowable_error = %g: allowable error limit\n", ALLOWABLE_ERROR);
    fprintf(stderr, "         epoch = %d: epoch \n", EPOCH);
    fprintf(stderr, "         sigma = %g: gaussian dispersion \n", SIGMA);
    fprintf(stderr, "         slope = %g: slope of sigmoid curve\n", SIGMOID_SLOPE);
    fprintf(stderr, "         to = %d: max iterations\n", TO);
    fprintf(stderr, "         rand_range: = %g: range for initial values\n", RAND_RANGE);
    fprintf(stderr, "         seed: seed of random number generator\n");
}

/* Parse command line arguments and set the global variables
 * The variables to be set are 
 * REGRESSION(default) or CLASSIFICATION, N_of_top, N_of_second,
 * RAND_RANGE, LEARNING_RATIO, SIGMOID_SLOPE, ERROR_LIMIT, ALLOWABLE_ERROR,
 * EPOCH, TO, seed, INPUT_FILE_NAME, TEACHER_FILE_NAME, INIT_FILE_NAME 
 */
int ParseArgs(int argc, char **argv)
{
    int argn=1;

    while ( argn < argc ) {
        if ((strcmp(argv[argn],"--help") == 0) || (strcmp(argv[argn],"-h") == 0)) {
            usage(argv[0]);     exit (EXIT_FAILURE);
        }
	if (strcmp(argv[argn], "-classification") == 0) {
	    CLASSIFICATION = YES;
	    REGRESSION = NO;
	}
	if (strcmp(argv[argn], "-regression") == 0) {
	    CLASSIFICATION = NO;
	    REGRESSION = YES;
	}
        if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
            if (strcmp(argv[argn], "-learn") == 0) {
                LEARNING_RATIO = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-rand_range") == 0) {
                RAND_RANGE = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-sigma") == 0) {
                SIGMA = atof(argv[++argn]);
		if ((SIGMA - 0.0) < eps) {
		    fprintf(stderr,"Too small sigma\n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-slope") == 0) {
                SIGMOID_SLOPE = atof(argv[++argn]);
            }
            if (strcmp(argv[argn],"-seed") == 0) {
                SEED = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-error") == 0) {
                ERROR_LIMIT = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-epoch") == 0) {
                EPOCH = atoi(argv[++argn]);
            }
	    if (strcmp(argv[argn], "-allowable_error") == 0) { 
		ALLOWABLE_ERROR = atof(argv[++argn]);
	    }
            if (strcmp(argv[argn], "-top") == 0) {
                N_of_top = atoi(argv[++argn]);
		if ( N_of_top < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "Invalid -top parameter\n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-second") == 0) {
                N_of_second = atoi(argv[++argn]);
		if ( N_of_second < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "Invalid -second parameter\n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-to") == 0) {
                TO = atoi(argv[++argn]);
		if ( TO < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "Invalid -to parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-input") == 0) {
		INPUT_FILE_NAME = (char *)malloc(strlen(argv[++argn])+1);
		if (INPUT_FILE_NAME == NULL) {
		    fprintf(stderr, "Could not allocate memory for INPUT_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(INPUT_FILE_NAME, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy INPUT_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		INPUT_FILE_NAME[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-teacher") == 0) {
		TEACHER_FILE_NAME = (char *)malloc(strlen(argv[++argn])+1);
		if (TEACHER_FILE_NAME == NULL) {
		    fprintf(stderr, "Could not allocate memory for TEACHER_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(TEACHER_FILE_NAME, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy TEACHER_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		TEACHER_FILE_NAME[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-init") == 0) {
		INIT_FILE_NAME = (char *)malloc(strlen(argv[++argn])+1);
		if (INIT_FILE_NAME == NULL) {
		    fprintf(stderr, "Could not allocate memory for INIT_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(INIT_FILE_NAME, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy INIT_FILE_NAME\n");
		    exit (EXIT_FAILURE);
		}
		INIT_FILE_NAME[strlen(argv[argn])] = '\0';
	    }
	}
	argn++;
    }

    if (INPUT_FILE_NAME == NULL) {
	usage(argv[0]);
	fprintf(stderr, "Invalid input_file_name=%s\n", INPUT_FILE_NAME);
	exit (EXIT_FAILURE);
    }
    if (TEACHER_FILE_NAME == NULL) {
	usage(argv[0]);
	fprintf(stderr, "Invalid filename of teacher=%s\n", TEACHER_FILE_NAME);
	exit (EXIT_FAILURE);
    }
    Input_Info = matinfo_read(INPUT_FILE_NAME);
    Teacher_Info = matinfo_read(TEACHER_FILE_NAME);

    N_of_input = Input_Info->m;
    if ( N_of_input <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    N_of_output = Teacher_Info->m;
    if ( N_of_output <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    Npats = Input_Info->n;
    Example = Input_Info->mat;
    Teacher = Teacher_Info->mat;

    return OK;
}

/* McMalloch and Pitts' formal neuron */ 
void formal_f( neuron *a )
{
    if ( a->state > 0 )
	a->output = 1.0;
    else
	a->output = 0.0;
}

/* so-called sigmoid function */
void logistic_f( neuron *a )
{
    a->output = 1.0 / ( 1.0 + exp( - SIGMOID_SLOPE * a->state ));
}

/* linear output function */
void linear_f( neuron *a )
{
    a->output = a->state;
}

/* Initialzing a neuron defined as a struct EN_NEURON */
void initialize_a_neuron( neuron *a, int Ninp, void output_func(neuron *) )
{
    int i;

    a->output = 0.0;
    a->state = 0.0;
    a->prior = 0.0;
    a->posterior = 0.0;
    a->output_func = output_func;
    a->Ninp = Ninp;

    if ( a->Ninp == 0 ) {
        a->inp_wgt = NULL;
        a->inp_neuron = NULL;
        return;
    }

    /* 0-th neuron will be handled as a threshold value */
    a->inp_wgt = (double *)malloc(sizeof(double) * (a->Ninp +1));
    if ( a->inp_wgt == NULL){
        fprintf(stderr, "Could not allocate memory for inp_wgt\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i<= a->Ninp; i++) {
        a->inp_wgt[i] = 2.0 * MTgenrand_f() * RAND_RANGE - RAND_RANGE;
    }

    /* 0-th neuron will be handled as a threshold value */
    a->inp_neuron = (neuron **)malloc(sizeof(neuron *) * (a->Ninp +1));
    if ( a->inp_neuron == NULL ) {
        fprintf(stderr, "Could not allocate memory for inp_neuron\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i <= a->Ninp; i++) {
        a->inp_neuron[i] = NULL;
    }
}

/* print information about a neuron to stderr */
void print_a_neuron( neuron *a )
{
    int i;

    fprintf(stderr, "\tAddres of neuron=%p", a );
    fprintf(stderr, "\toutput=%f\n", a->output);
    fprintf(stderr, "\tstate=%f\n", a->state);
    fprintf(stderr, "\tprior=%f\n", a->prior);
    fprintf(stderr, "\tposterior=%f\n", a->posterior);
    fprintf(stderr, "\tNinp   :%d\n", a->Ninp);
    for (i=0; i< a->Ninp; i++) {
	fprintf(stderr, "\tinp_wgt[%d]=%5.3f(%p)\n", i, a->inp_wgt[i], &a->inp_wgt[i]);
	fprintf(stderr, "\tinp_neuron[%d].output=%g(%p)\n", i, a->inp_neuron[i]->output, a->inp_neuron[i]);
    }
}


void build_experts_network( void )
{
    int i,j,k,l;

    experts = (EN **)malloc(sizeof(EN *) * (N_of_top));
    if ( experts == NULL ) {
	fprintf(stderr, "Could not allocate memroy for *experts\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i< N_of_top; i++) {
        experts[i] = (EN *)malloc(sizeof(EN) * (N_of_second));
	if ( experts[i] == NULL ) {
	    fprintf(stderr, "Could not allocate memroy for experts[%d]\n",i);
	    exit (EXIT_FAILURE);
	}
        for (j=0; j<N_of_second; j++) {
	    experts[i][j].vector = (neuron *)malloc(sizeof(neuron) * (N_of_output));
	    if ( experts[i][j].vector == NULL ) {
		fprintf(stderr, "Could not allocate memroy for experts[%d][%d].vector\n",i,j);
		exit (EXIT_FAILURE);
	    }
	    for (k=0; k < N_of_output; k++) {
		if ( REGRESSION ) {
		    initialize_a_neuron(&experts[i][j].vector[k], N_of_input, linear_f);
		} else {
		    initialize_a_neuron(&experts[i][j].vector[k], N_of_input, logistic_f);
		}
		for (l=0; l<=N_of_input; l++) {
		    experts[i][j].vector[k].inp_neuron[l] = &input_vector[l];
		}
	    }
	}
    }

    Uij = (EN **)malloc(sizeof(EN *) * (N_of_top));
    if ( Uij == NULL ) {
	fprintf(stderr, "Could not allocate memroy for Uij\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<N_of_top; i++) {
	Uij[i] = (EN *)malloc(sizeof(EN) * (N_of_second));
	if ( Uij[i] == NULL ) {
	    fprintf(stderr, "Could not allocate memory for Uij[%d]\n", i);
	    exit (EXIT_FAILURE);
	}
	for (j=0; j<N_of_second; j++) {
	    Uij[i][j].vector = (neuron *)malloc(sizeof(neuron) * (N_of_output));
	    if ( Uij[i][j].vector == NULL ) {
		fprintf(stderr, "Could not allocate memroy for Uij[%d][%d]\n",i,j);
		exit (EXIT_FAILURE);
	    }
	    for (k=0; k<N_of_output; k++) {
		initialize_a_neuron(&Uij[i][j].vector[k], 0, NULL);
	    }
	}
    }
}

void build_gating_network( void )
{
    int i,j,k;

    G1.member = (neuron *)malloc(sizeof(neuron) * (N_of_top));
    if ( G1.member == NULL ) {
	fprintf(stderr, "Could not allocate memory for G1.member\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<N_of_top; i++) {
	initialize_a_neuron(&G1.member[i], N_of_input, linear_f);
	for ( j=0; j<=N_of_input; j++ ) {
	    G1.member[i].inp_neuron[j] = &input_vector[j];
	}
    }

    G2 = (GN *)malloc(sizeof(GN) * (N_of_top));
    if ( G2 == NULL ) {
	fprintf(stderr, "Could not allocate memory for G2\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<N_of_top; i++) {
	G2[i].member = (neuron *)malloc(sizeof(neuron) * (N_of_second));
	if ( G2[i].member == NULL ) {
	    fprintf(stderr, "Could not allocate memory for G2[%d].member\n", i);
	    exit (EXIT_FAILURE);
	}
	for (j=0; j<N_of_second; j++) {
	    initialize_a_neuron(&G2[i].member[j], N_of_input, linear_f);
	    for (k=0; k<=N_of_input; k++) {
		G2[i].member[j].inp_neuron[k] = &input_vector[k];
	    }
	}
    }
}

void read_weights_from_init_file( void )
{
    int i, j, k, l;
    FILE *fp;
    matrix TMP;

    TMP = matinit(1,N_of_input+1);
    fp = fileio_util_openr(INIT_FILE_NAME);

    for (i=0; i<N_of_top; i++) {
	freadmat(fp, TMP, 1, N_of_input+1);
	for ( j=0; j<=N_of_input; j++) {
	    G1.member[i].inp_wgt[j] = *(TMP+offs(1,N_of_input+1,1,j+1));
	}
    }

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    freadmat(fp, TMP, 1, N_of_input+1);
	    for (k=0; k<=N_of_input; k++) {
		G2[i].member[j].inp_wgt[k] = *(TMP+offs(1,N_of_input+1,1,k+1));
	    }
	}
    }

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<N_of_output; k++) {
		freadmat(fp, TMP, 1, N_of_input+1);
		for (l=0; l<=N_of_input; l++) {
		    experts[i][j].vector[k].inp_wgt[l] = *(TMP+offs(1,N_of_input+1,1,l+1));
		}
	    }
	}
    }
    fileio_util_close(fp);
    free (TMP);
}


void build_network( void ) 
{
    int i;

    if ( ERROR_LIMIT == 0.0 ) {
	ERROR_LIMIT = 0.05;
    }

    /* 0-th neuron will be handled as a threshold unit */
    input_vector = (neuron *)malloc(sizeof(neuron) * (N_of_input+1));
    if ( input_vector == NULL ) {
	fprintf(stderr, "Could not allocate memory for input vectorn");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<=N_of_input; i++) {
	initialize_a_neuron(&input_vector[i], 0, NULL);
    }

    output_vector = (neuron *)malloc(sizeof(neuron) * (N_of_output));
    if ( output_vector == NULL ) {
	fprintf(stderr, "Could not allocate memroy for output_vector\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<N_of_output; i++) {
	initialize_a_neuron(&output_vector[i], 0, NULL);
    }

    Pij = matinit(N_of_top, N_of_second);
	
    build_gating_network();
    build_experts_network();

    /* When INIT_FILE_NAME will be given, then          */
    /* read the initial value of weight value from file */
    if ( INIT_FILE_NAME != NULL ) {
	read_weights_from_init_file();
    }
}

/* calc an output correspondings to an input vector */
void calc_an_output( neuron *a )
{
    int i;
    double wrk = 0.0;
    static int t=0;

    t++;
    for (i=0; i<=a->Ninp; i++) {
	wrk += a->inp_wgt[i] * a->inp_neuron[i]->output;
    }
    a->state = wrk;
    a->output_func( a );
}


/* Set values for an input vector to  each input neuron */
/* where pattern number will be given by second argument pat(int) */
void set_input( int pat )
{
    int i;

    /* 0-th neuron will be handled as a threshold unit */
    input_vector[0].output = 1.0;
    for (i=1; i<=N_of_input; i++) {
	input_vector[i].output = *(Example+offs(Npats,N_of_input,pat+1,i));
    }
}

/* calc probability $P_{ij}(y|x,\theta)$ */
double calc_Pij( EN a, int pat )
{
    int i;
    double wrk, y, u;
    double mean, s1, ss, n;

    wrk = 0.0;
    if ( REGRESSION ) { /* regression case */
	mean = 0.0; s1 = 0.0; ss = 0.0; n = 0.0;
	for(i=0; i<N_of_output; i++){
	    /* Gaussian distribution with identity covariance matrix */
	    y = *(Teacher+offs(Npats,N_of_output,pat+1,i+1));
	    u = a.vector[i].output;
	    wrk += (y-u) * (y-u);
            n += 1.0;
            s1 = y - mean;
            mean += s1 / n;
            ss += ( n - 1.0) / n * s1 * s1;
	}
	ss = ss / n;
	if ( ss <= 0.001 ) {
		ss = 1.0;
	}
	return exp(- 1.0 / 2.0 * wrk / (ss * SIGMA) );
    } else { /* classification case */
	for(i=0; i<N_of_output; i++){
	    /* multinomial Bernoulli distribution */
	    y = *(Teacher+offs(Npats,N_of_output,pat+1,i+1));
	    u = a.vector[i].output;
	    wrk += pow(u,y) * pow((1.0-u),(1.0-y));
	}
	return ( wrk );
    }
}

/* calc softmax function, which will be called from calc_output_vector() */
void softmax( GN *gates, int n )
{
    int i;
    double wrk;

    wrk = 0.0;
    for (i=0; i<n; i++) {
	wrk += exp(gates->member[i].output);
	if ( isnan(wrk) != 0 ) {
	    fprintf(stderr, "### parameter too huge in softmax\n");
	    fprintf(stderr, "### gates->member[%d].output=%f\n",
		    i, gates->member[i].output);
	    exit (EXIT_FAILURE);
	}
    }
    for (i=0; i<n; i++) {
	gates->member[i].prior = exp(gates->member[i].output) / wrk;
	if ( isnan(gates->member[i].prior) != 0 ) {
	    fprintf(stderr, "### gates->member[%d].prior=%f wrk=%f in softmax()\n",
		    i,gates->member[i].prior, wrk);
	    exit (EXIT_FAILURE);
	}
    }
}

void calc_output_vector( int pat )
{
    int i,j,k;

    set_input( pat );

    for (i=0; i<N_of_top; i++) {
	calc_an_output(&G1.member[i]);
	if ( isnan(G1.member[i].output) != 0 ) {
	    fprintf(stderr,"### G1.member[%d].output=%f in calc_output_vector()\n",
		i, G1.member[i].output);
	    exit (EXIT_FAILURE);
	}
        for (j=0; j<N_of_second; j++) {
	    calc_an_output(&G2[i].member[j]);
	    if ( isnan(G2[i].member[j].output) != 0 ) {
		fprintf(stderr, "### G2[%d].member[%d].output=%f\n", 
			i, j, G2[i].member[j].output);
		exit (EXIT_FAILURE);
	    }
	    for (k=0; k<N_of_output; k++) {
		calc_an_output(&experts[i][j].vector[k]);
#ifdef DEBUG
		if (CLASSIFICATION) {
		    if (experts[i][j].vector[k].output < 0.0) {
			printf("Strange in calc_output_vector()\n");
			exit(EXIT_FAILURE);
		    }
		}
#endif
		if ( isnan(experts[i][j].vector[k].output) != 0 ) {
		    fprintf(stderr, "### experts[%d][%d].vector[%d]=%f\n",
			    i, j, k, experts[i][j].vector[k].output);
		    exit (EXIT_FAILURE);
		}
	    }
	    
	    *(Pij+offs(N_of_top,N_of_second,i+1,j+1)) = calc_Pij(experts[i][j], pat);
	    if ( isnan( *(Pij+offs(N_of_top, N_of_second, i+1, j+1)) ) != 0 ) {
		fprintf(stderr, "### Pij=%f,(i,j)=(%d,%d)\n",
			*(Pij+offs(N_of_top, N_of_second, i+1, j+1)), i, j); 
		exit (EXIT_FAILURE);
	    }
	}
    }

    /* calc softmax for the top level gating network */
    softmax(&G1, N_of_top);

    /* calc softmax for the second level gating network */
    for (i=0; i< N_of_top; i++ ) {
	softmax(&G2[i], N_of_second);
    }

    /* calc second level output vector */
    for (i=0; i< N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<N_of_output; k++) {
		Uij[i][j].vector[k].output = 0.0;
	    }
	    for (k=0; k<N_of_output; k++) {
		/* store the ouput of node[i][j] to Uij[i][j] */
		/* $U_{ij} = g_{j|i} P_{ij}(y)$ */
		Uij[i][j].vector[k].output += experts[i][j].vector[k].output * G2[i].member[j].prior;
		if ( isnan(Uij[i][j].vector[k].output) != 0 ) {
		    fprintf(stderr, "### Uij[%d][%d].vector[%d].output=%f in calc_output_vector()\n",i,j,k, Uij[i][j].vector[k].output);
		    exit (EXIT_FAILURE);
		}
	    }
	}
    }

    for (k=0; k<N_of_output; k++) {
	output_vector[k].output = 0.0;
    }

    /* calc top level output vector */
    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<N_of_output; k++) {
		output_vector[k].output += Uij[i][j].vector[k].output * G1.member[i].prior;
		if ( isnan( output_vector[k].output) != 0 ) {
		    fprintf(stderr, "### output_vector[%d]=%f in calc_output_vector\n",
			    k, output_vector[k].output);
		    exit (EXIT_FAILURE);
		}
	    }
	}
    }
}

void calc_posterior( void )
{
    int i,j;
    double wrk, denominator;

    /* second level posterior probability $h_{j|i}$ */
    for (i=0; i< N_of_top; i++) {
	wrk = 0.0;
	for (j=0; j<N_of_second; j++) {
	    /* $P_{ij} * g_{j|i}$ */
	    wrk += G2[i].member[j].prior * *(Pij+offs(N_of_top,N_of_second,i+1,j+1));
	}
	denominator = wrk; /* $\sum_j g_{j|i} Pij(y)$ */
	for (j=0; j<N_of_second; j++) {
	    wrk = G2[i].member[j].prior * *(Pij+offs(N_of_top,N_of_second,i+1,j+1));
	    G2[i].member[j].posterior = wrk/denominator;
	}
    }

    /* top level posterior probability $h_i$ */
    denominator = 0.0;
    for (i=0; i<N_of_top; i++) {
	wrk=0.0;
	for (j=0; j<N_of_second; j++) {
	    wrk += G2[i].member[j].prior * *(Pij+offs(N_of_top,N_of_second,i+1,j+1));
	}
	/* $ \sum_j g_j \sum_k g_{k|j} Pjk(y)$ */		
	denominator += G1.member[i].prior * wrk; 
    }
    for (i=0; i<N_of_top; i++) {
	wrk = 0.0;
	for (j=0; j<N_of_second; j++) {
	    wrk += G2[i].member[j].prior * *(Pij+offs(N_of_top,N_of_second,i+1,j+1));
	}
	G1.member[i].posterior = (G1.member[i].prior * wrk) / denominator;
    }
}

/* calc Sum of Square of Errors */
double calc_SSE( void )
{
    int i,pat;
    double error=0.0, ret=0.0;

    for (pat=0; pat<Npats; pat++) {
	calc_output_vector( pat );
	calc_posterior();
	for (i=0; i<N_of_output; i++) {
	    error = (output_vector[i].output 
		     - *(Teacher+offs(Npats,N_of_output,pat+1,i+1)));
	    if (fabs(error) < ALLOWABLE_ERROR) { error = 0.0; }
	    ret += error * error;
	}
    }
    return ret;
}

/* Print an output correspondings to a given pattern */
void print_an_output( int pat )
{
    int i, j, k;

    fprintf(stderr,"# Pattern %d: (", pat);
    for (i=1; i<=N_of_input; i++) {
	fprintf(stderr," %f", input_vector[i].output);
    }
    fprintf(stderr,")->output(");
    for (i=0; i<N_of_output; i++) {
	fprintf(stderr," %f", output_vector[i].output);
    }
    fprintf(stderr,")=teacher(");
    for (i=1; i<=N_of_output; i++) {
	fprintf(stderr," %f", *(Teacher+offs(Npats,N_of_output,pat+1,i)));
    }
    fprintf(stderr,"), gating network(");
    for (i=0; i<N_of_top; i++) {
	fprintf(stderr," %f", G1.member[i].prior);
    }
    fprintf(stderr,"), second gating network(");
    for (i=0; i<N_of_top; i++) {
	fprintf(stderr,"(");
	for (j=0; j<N_of_second; j++) {
	    fprintf(stderr," %f", G2[i].member[j].prior);
	}
	fprintf(stderr,")");
    }
    fprintf(stderr,"), output of each experts(");
    for (i=0; i<N_of_top; i++) {
	fprintf(stderr,"(");
	for (j=0; j<N_of_second; j++) {
	    fprintf(stderr,"(");
	    for (k=0; k<N_of_output; k++) {
		fprintf(stderr," %f", experts[i][j].vector[k].output);
#ifdef DEBUG
		if ( CLASSIFICATION ) {
		    if ( experts[i][j].vector[k].output < 0.0 ) {
			printf("Strange in print_an_output");
			print_a_neuron( &experts[i][j].vector[k] );
			exit (EXIT_FAILURE);
		    }
		}
#endif
	    }
	    fprintf(stderr,")");
	}
    }
    fprintf(stderr,")\n");
}


/* Print results to the stderr */
void print_results( void )
{
    int pat;

    for (pat=0; pat<Npats; pat++) {
	calc_output_vector( pat );
	calc_posterior();
	print_an_output(pat);
    }
}

/* Print weight matrix to the stdout */
void print_weights( void )
{
    int i,j,k,l;

    for (i=0; i<N_of_top; i++) {
	fprintf(stdout,"###  G1.member[%d]\n", i);
	for (j=0; j<=N_of_input; j++) {
	    fprintf(stdout," %+-9.6f ", G1.member[i].inp_wgt[j]);
	}
	fprintf(stdout,"\n");
    }

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    fprintf(stdout, "### G2[%d].member[%d]\n",i,j);
	    for (k=0; k<=N_of_input; k++) {
		fprintf(stdout, " %+-9.6f ", G2[i].member[j].inp_wgt[k]);
	    }
	    fprintf(stdout,"\n");
	}
    }

    fprintf(stdout,"### experts\n");
    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<N_of_output; k++) {
		fprintf(stdout, "### experts[%d][%d].vector[%d]\n",i,j,k);
		for (l=0; l<=N_of_input; l++) {
		    fprintf(stdout, " %+-9.6f ", experts[i][j].vector[k].inp_wgt[l]);
		}
		fprintf(stdout,"\n");
	    }
	}
    }
}

/* genrate a randomized sequence for stochastic learning */
void do_shuffle( int *order, int n )
{
    int i, index, tmp;

    for (i=0; i<n; i++) {
        index = i + (int)(MTgenrand_f() * ((double)n - (double)i ));
	tmp = order[index];
	order[index] = order[i];
	order[i] = tmp;
    }
}

/* learning of top gating network */
/* $v_i += \rho ( h_i - g_i ) x^t$ */
void update_G1( void )
{
    int i,j;

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<=N_of_input; j++) {
	    G1.member[i].inp_wgt[j]
		+= LEARNING_RATIO 
		* ( G1.member[i].posterior - G1.member[i].prior ) 
		* input_vector[j].output;
	}
    }
}

/* learning of second gating network */
/* $v_{ij} += \rho h_i ( h_{j|i} - g_{j|i} ) x^t$ */
void update_G2( void )
{
    int i,j,k;

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<=N_of_input; k++) {
		G2[i].member[j].inp_wgt[k]
		    += LEARNING_RATIO
		    * G1.member[i].posterior
		    * ( G2[i].member[j].posterior - G2[i].member[j].prior )
		    * input_vector[k].output;
	    }
	}
    }
}

/* learning of experts network */
/* $U_{ij} += \rho h_i h_{j|i} ( y - u ) x^t$ */
void update_EN( int pat )
{
    int i,j,k,l;

    for (i=0; i<N_of_top; i++) {
	for (j=0; j<N_of_second; j++) {
	    for (k=0; k<N_of_output; k++) {
		for (l=0; l<=N_of_input; l++) {
		    experts[i][j].vector[k].inp_wgt[l] 
			+= LEARNING_RATIO
			* G1.member[i].posterior
			* G2[i].member[j].posterior
			* ( *(Teacher+offs(Npats,N_of_output,pat+1,k+1)) - experts[i][j].vector[k].output )
			* input_vector[l].output;
		}
	    }
	}
    }
}

/* main loop of computation */
int main_loop( double *SSE )
{
    int t, i, pat;
    int *shuffle;
    double mse;

    shuffle = (int *)malloc(sizeof(int) * Npats);
    if ( shuffle == NULL ) {
	fprintf(stderr, "Could not allocate memory for shuffle\n");
	exit (EXIT_FAILURE);
    }
    for (i=0; i<Npats; i++) {
	shuffle[i] = i;
    }

    for (t=0; t<TO; t++) { /* main iteration t */
	*SSE = calc_SSE();
	mse = *SSE / ( (double)Npats * (double)N_of_output );
	mse = sqrt(mse);
	if ( mse <= ERROR_LIMIT) {
	    return t;
	}
	if ( (t % EPOCH) == 0 ){
	    fprintf(stderr,"%05d %g ### MSE\n", t, mse);
	}
	do_shuffle( shuffle, Npats );
	for (pat=0; pat<Npats; pat++){ 
	    calc_output_vector( shuffle[pat] );
	    calc_posterior();
	    update_G1();
	    update_G2();
	    update_EN( shuffle[pat] );

	} /* for each pattern */
    } /* for iteration t */
    return t;
}

int main( int argc, char **argv )
{
    int iterations = 0;
    double SSE = 0.0;
    double mse = 0.0;

    ParseArgs(argc,argv);
    init_random(SEED);

    build_network();

    iterations = main_loop( &SSE );
    print_results();
    print_weights();

    fprintf(stdout, "### %s %s -input %s -teacher %s %s %s -top %d -second %d -learn %g -error %g -sigma %g -slope %g -rand_range %g -epoch %d -to %d -allowable_error %g -seed %d\n",
	    argv[0], 
	    (REGRESSION==YES) ? "-regression" : "-classification",
	    INPUT_FILE_NAME, TEACHER_FILE_NAME, 
            (INIT_FILE_NAME != NULL) ? "-init" : "",
            (INIT_FILE_NAME != NULL) ? INIT_FILE_NAME : "",
	    N_of_top, N_of_second, 
	    LEARNING_RATIO, ERROR_LIMIT, SIGMA, SIGMOID_SLOPE, 
	    RAND_RANGE, EPOCH, TO, ALLOWABLE_ERROR, get_seed() );

    mse = SSE / ((double)Npats * (double)N_of_output);
    mse = sqrt(mse);
    if ( iterations == TO ) {
	fprintf(stdout,"### COULD NOT LEARN AT ALL WITHIN %d TIMES MSE=%g\n", 
		iterations, mse);
    } else {
	fprintf(stdout,"### Learning completed in %d time(s) MSE=%g\n", 
		iterations, mse);
    }
    return OK;
}
