/*
 * @(#)freadmat.c
 * by Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 */
#include <stdlib.h>
#include <string.h>
#include "general.h"
#include "matrix_util.h"
#include "fileio_util.h"

matrix matreadfile(char *filename, int n, int m)
{
    FILE *fp;
    matrix ret;

    if ( n <= 0 ) {
        fprintf(stderr, "Invalid size n(%d) in matreadfile()\n", n);
        exit (EXIT_FAILURE);
    }
    if ( m <= 0 ) {
        fprintf(stderr, "Invalid size m(%d) in matreadfile()\n", m);
        exit (EXIT_FAILURE);
    }
    ret = matinit(n,m);

    fp = fileio_util_openr(filename);
    freadmat(fp, ret, n, m); 
    fileio_util_close(fp);
    return ret;
}

int vectread(FILE *fp, matrix w, int size) 
{
    return freadmat(fp, w, size, 1);
}
int freadmat(FILE *fp, matrix A, int on, int om)
{
    char *line, *cp, *token;
    scalar ent=0.0;
    int n = 0, m = 0;

    if ( A == NULL ) {
        fileio_util_perror( "Invalid matrix A in freadmat()\n"); 
        exit (EXIT_FAILURE);
    }
    
    while ((line=getline(fp)) != NULL) {
        if (line == NULL) {
            fileio_util_perror("read error in freadmat()\n");
            return NO;
        }
        
        /* comment out '#' skip*/
        cp = index(line, '#');
        if (cp != NULL) {
            fprintf(stdout, "%s\n", cp);
            *cp = '\0';
        }
        if (line[0]=='\0') continue;
        
        n++;
        if ((token = strtok(line, SEPARATOR_CHARS)) == NULL){
            fprintf(stderr, "Can't read in readmat(n=%d,m=%d)\n", n+1, m);
            exit (EXIT_FAILURE);
        }
        m=1;
/**********************************************************/
/* if scalar which defined in matrix_util.h is double,
 "%lg" argument have to be used.   On the other hand,
 "%g" must be used in case that you define scalar as float */
/**********************************************************/
#ifdef SCALAR_AS_FLOAT
        if (sscanf(token, "%g", &ent) == 0 ) {
#else
        if (sscanf(token, "%lg", &ent) == 0 ) {
#endif
             fprintf(stderr, "Could not assign a value in freadmat()\n");
             exit (EXIT_FAILURE);
        }
        *(A+offs(on,om,n,m)) = ent;
        if ((n==on) && (m==om)) return OK;
        while ( (token = strtok(NULL,SEPARATOR_CHARS)) != NULL) {
#ifdef SCALAR_AS_FLOAT
            if (sscanf(token, "%g", &ent) <= 0) {
#else
            if (sscanf(token, "%lg", &ent) <= 0) {
#endif
                fprintf(stderr, "readmat:maybe not a number(n=%d,m=%d)\r",n+1,m+1);
                break;
            }
            m++;
            *(A+offs(on,om,n,m)) = ent;
            if ((n==on) && (m==om)) return OK;
        }
        if (om != m) {
            fprintf(stderr,"lack of data in n=%d,m=%d(on=%d,om=%d)\n",n,m,on,om);
            exit (EXIT_FAILURE);
        }
    }
    if (on != n){
        fprintf(stderr,"lack of lines in n=%d,m=%d\n", n, m);
        exit (EXIT_FAILURE);
    }
    if ((n!=on) || (m!=om)) { 
        fprintf(stderr,"Could not read data file in freadmat()");
        exit (EXIT_FAILURE);
    }
    return OK;
}

#ifdef TEST
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>

int usage(char *prgname)
{
    printf("%s  - test of reading a matrix\n", prgname);
    printf("Usage %s options [datafile]\n", prgname);
    printf("options: \n");
    printf(" -n rows : size of row of datafile\n");
    printf(" -m cols : size of column of datafile\n");
    printf(" -help : show this help\n");
    exit (EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    int   n=0,m=0, argn = 1;
    matrix A;
    FILE *fp;

    fprintf(stderr,"argc=%d\n",argc);

    while (argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0') {
        if (keymatch(argv[argn], "--help")) {
            usage(argv[0]);
        } else if (keymatch(argv[argn],"-h")) {
            usage(argv[0]);
        } else if (keymatch(argv[argn],"-n")) {
            ++argn;
            if ((argn==argc) || (sscanf(argv[argn],"%d",&n)!=1))
                usage(argv[0]);
        } else if (keymatch(argv[argn],"-m")) {
            ++argn;
            if ((argn==argc) || (sscanf(argv[argn],"%d",&m)!=1))
                usage(argv[0]);
        } else { usage(argv[0]); }
        ++argn;
    }
  
    if ((n == 0) || (m == 0)) usage(argv[0]);
    A  = matinit(n,m);

    if (argn != argc) {
        fp = fileio_util_openr(argv[argn]);
        ++argn;
    } else {
        fp = stdin;
    }
    if (argn != argc)
        usage(argv[0]);
    freadmat (fp, A, n, m);
    fileio_util_close(fp);
    fwritemat(stdout,A,n,m,"Result:\n","%g ");
    return OK;
}
#endif TEST
