/* Read the data file (matrix format), assign each value to a matrix,
   which is a member of struct matinfo defined in matrix_util.h 

   Copyright (c) 1999 Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 
   NOTE: 
   A data file must be consisted of several lines,
   and each line must be consisted of several columns.
   Each column must be separated by one of SEPARATOR_CHARS.
   The SEPARATOR_CHARS is defined in matrix_util.h

   Disclaimer:  No guarantees of performance accompany this software,
   nor is any responsibility assumed on the author.  */
#include <strings.h>
#include <stdlib.h>
#include <string.h>
#include "matrix_util.h"
#include "fileio_util.h"

/* colum_list(column_info) is a list structure, consisted of a value
   and a pointer to the next data. */
typedef struct column_info {
    struct column_info *next;
    scalar value;
} column_list;

/* line_list is a list structure, consisted of a colum_list and a
   pointer indicated to the next line. */
typedef struct line_info {
    struct line_info *next;
    column_list *cols;
} line_list;

/* Allocate some amounts of memories required for one column */
column_list *alloc_col(void)
{
    column_list *col;

    col = (column_list *)malloc(sizeof(column_list));
    if (col == NULL) {
        fileio_util_perror("Could not allocate memory in alloc_col()");
        exit (EXIT_FAILURE);
    }
    col->value = 0.0;
    col->next = NULL;
#ifdef DEBUG2
    fprintf(stderr, "\talloc_col address: %x\n", (int)col);
#endif
    return col;
}

/* Allocate some amounts of memories required for one line */
line_list *alloc_line(void)
{
    line_list *line;
    
    line = (line_list *)malloc(sizeof(line_list));
    if (line == NULL) {
        fileio_util_perror("Could not allocate memory in alloc_line()");
        exit (EXIT_FAILURE);
    }
    line->cols = NULL;
    line->next = NULL;
#ifdef DEBUG2
    fprintf(stderr, "alloc_line address: %x\n", (int)line);
#endif
    return line;
}

/* free memory for a column_list */
void free_column(column_list *col, int count)
{
#ifdef DEBUG2
    fprintf(stderr,"(%d) start of  free_column()\n", count);
#endif
    if (col->next == NULL) {
#ifdef DEBUG2
        fprintf(stderr,"(%d)col->next is NULL free_column(), address=%x\n", count,(int)col);
#endif
	free(col);
        return;
    } else {
        free_column(col->next, count + 1);
#ifdef DEBUG2
        fprintf(stderr, "\taddres of col=%x in free_column(), where value=%g\n", (int)col, col->value);
#endif
        return;
    }
}

/* free memory for a line_list */
void free_line(line_list *line, int count)
{
#ifdef DEBUG2
    fprintf(stderr, "95 Start of free_line() address of line_lists=%x count=%d\n", (int)line, count);
#endif
    if (line->next == NULL) { /* last entry of the input line */
#ifdef DEBUG2
        fprintf(stderr,"99 line->next is NULL (%d) free_line()\n" , count);
#endif
	if ( line->cols != NULL) {
#ifdef DEBUG2
	    fprintf(stderr,"103 line->cols==NULL then return (%d)\n" , count);
#endif
	    free_column(line->cols, 1);
	}
#ifdef DEBUG2
	fprintf(stderr, "108 address of line=%x at free_line at count=%d\n", (int)line, count);
#endif
        return;
    } else {
#ifdef DEBUG2
        fprintf(stderr, "113 (%d)address of line=%x at free_line()\n", count,(int)line);
#endif
        free_line(line->next, count + 1);
#ifdef DEBUG2
        fprintf(stderr, "117(%d)address of line=%x at free_line()\n", count,(int)line);
#endif
	if ( count != 1 ) {
#ifdef DEBUG2
	    fprintf(stderr, "HIT COUNT=1 address of line=%x at free_line()\n", (int)line);
#endif
	    free(line);
	}
    }
}

/* Analyze an input line, decompose it to several values, 
 * and make column_list, then return this column_list(struct defined above)
 * This function will be called from lineinfo_readfile() described below
 */
column_list *make_columnlist(char *line, int *m)
{
    char *token;
    scalar ent;
    column_list *ret, *cur, *next_col;
    
    ret = alloc_col();
#ifdef DEBUG2
    fprintf(stderr, "address of ret=%x at make_columnlist()\n", (int)ret);
#endif
    if ( (token = strtok(line, SEPARATOR_CHARS)) == NULL) {
        fileio_util_perror("Could not strtok at make_columnlist()");
        exit (EXIT_FAILURE);
    }
#ifdef SCALAR_AS_FLOAT
    if (sscanf(token,"%g", &ent) == 0) {
#else
    if (sscanf(token,"%lg", &ent) == 0) {
#endif
        fileio_util_perror("Illeagal character at make_columnlist()");
        exit (EXIT_FAILURE);
    }
    ret->value = ent;
    *m = 1;
    cur = ret;
    while ( (token = strtok(NULL, SEPARATOR_CHARS)) != NULL) {
#ifdef SCALAR_AS_FLOAT
        if (sscanf(token, "%g", &ent) <= 0) {
#else
        if (sscanf(token, "%lg", &ent) <= 0) {
#endif
            fprintf(stderr,"Illegal character(%s) at %d th line\n",token,*m);
            break;
        }
        next_col = alloc_col();
#ifdef DEBUG2
    fprintf(stderr, "address of next_col=%x at make_columnlist()\n", (int)next_col);
#endif
        next_col->value = ent;
        (*m)++;
        cur->next = next_col;
        cur = next_col;
    }
    return ret;
}


/* Read a data of matrix format, and return a line list(struct defined above)
 * Each value of a matrix will assingned by matinfo_read()
 */
line_list *lineinfo_readfile(char *filename, int *n, int *m)
{
    FILE *fp;
    line_list *startline, *cur, *next;
    char *a_line, *cp;
    int maxcol;
    
    *n = 0; *m = 0;
    startline = alloc_line();
    cur = startline;
    
    fp = fileio_util_openr(filename);
    while ( (a_line = getline(fp)) != NULL) {
        cp = index(a_line, '#');
        if (cp != NULL) {
            fprintf(stdout, "%s\n", cp);
            *cp = '\0';
        }
        if (a_line[0] == '\0') continue;
        
        cur->cols = make_columnlist(a_line, &maxcol);
        if (maxcol > *m) *m = maxcol;
        *n = *n + 1;
        next = alloc_line();
        cur->next = next;
        cur = next;
    }
#ifdef DEBUG2
    fprintf(stderr, "### last cur addres=%x, to be freed###\n", (int)cur);
#endif
    fileio_util_close(fp);
    if ((*n == 0) || (*m == 0)) {
        fileio_util_perror("read error at lineinfo_readfile()");
        exit (EXIT_FAILURE);
    }
    free(a_line);
    return startline;
}

/* read a data from the file which disignated by filename, 
 * and return matinfo(struct defined at matrix_util.h)
 */
matinfo matinfo_read(char *filename)
{
    int i,j;
    line_list *line_lists, *aline;
    column_list *cols;
    matinfo ret;
    int n = 0, m = 0;

    if ((ret = (matinfo)malloc(sizeof(matinfo)))==NULL) {
        fileio_util_perror("Could not allocate memory at matinfo_read()");
        exit (EXIT_FAILURE);
    }
    if ((ret->name = (char *)malloc(strlen(filename) + 1))==NULL) {
        fileio_util_perror("Could not allocate memory at matinfo_read()");
        exit (EXIT_FAILURE);
    }
    if (strcpy(ret->name, filename) == NULL) {
        fileio_util_perror("Could not copy filename at matinfo_read()");
        exit (EXIT_FAILURE);
    }
    *(ret->name + strlen(filename)) = '\0';

    line_lists = lineinfo_readfile(filename, &n, &m);
#ifdef DEBUG2
    fprintf(stderr, "address of line_lists=%x at matinfo_read()\n", (int)line_lists);
#endif
    ret->n = n; 
    ret->m = m;
    ret->mat = matinit(ret->n, ret->m);

    for (i=1, aline=line_lists; i<=n; i++, aline=aline->next) {
        for (j=1, cols=aline->cols; j<=m; j++, cols=cols->next) {
            if (isnan(cols->value) != 0) {
								fileio_util_perror("Invalid input");
                fprintf(stderr,"i=%d,j=%d, illegal input\n", i, j);
                continue;
            }
            *(ret->mat+offs(n,m,i,j)) = cols->value;
            if (cols->next == NULL) break;
        }
    }
    if ( line_lists != NULL ) {
	/* forwarding version of free() for lines and columns     */ 
	/* System 7 like SUN sparc can do this routine.           */
        /* But I do not recommend you for the sake of portability */
        /* If you want to use this route, uncomment the following */
	/* at you own risk.                                       */
	/*
        for (i=1, aline=line_lists; i<=n; i++, aline=aline->next)
            for (j=1, cols=aline->cols; j<=m; j++, cols=cols->next)
                free(acol);
            free(aline);
        }
	*/
        free_line(line_lists,1);
    }
    return ret;
}

#ifdef STANDALONE
#include <stdio.h> 
#include <stdlib.h>

void usage(char *prgname)
{
    fprintf(stderr,"%s - show matrix size(rows column)\n", prgname); 
    fprintf(stderr,"Usage : %s [datafile] ...\n",prgname); 
    exit (EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    int   argn = 1;
    matinfo A;

    if (argc == 1) {
        A = matinfo_read ("-");
        fprintf(stdout, "### ProgramName: %s -\n", argv[0]);
        fprint_signature(stdout);
        fprintf (stdout, "%d %d\n", A->n, A->m);
    } else {
    while (argn < argc) {
        if (keymatch(argv[argn], "--help")) { usage(argv[0]); } 
        else if (keymatch(argv[argn], "-h")) { usage(argv[0]); } 
        else {
            A = matinfo_read(argv[argn]);
            fprintf(stdout, "### ProgramName: %s %s\n", argv[0], A->name);
            fprint_signature(stdout);
            fprintf(stdout, "%d %d\n", A->n, A->m);
        }
        ++argn;
    }
		}
    return OK;
}
#endif STANDALONE
