/* 
 * A perceptron simulator
 *
 * ProgramName: perceptron.c 
 * Author: Shinichi ASAKAWA <asakawa@twcu.ac.jp>
 * $Id: perceptron.c,v 1.4 2002/11/23 13:08:27 asakawa Exp asakawa $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "general.h"
#include "rand_gen.h"
#include "fileio_util.h"
#include "matrix_util.h"

typedef struct NEURON {
    double output;              /* Output values of this neuron */
    int Ninp;                   /* Number of input neurons */
    double *inp_wgt;            /* array of input weights */
    struct NEURON **inp_neuron; /* the pointer to the input neurons */
    double (*output_func)();    /* the pointer to the output function */
    double (*Doutput_func)();   /* the pointer to the differenctial above */
    double error;
} neuron;

/* ------------------------------------------------------------------------
   initial set of global variables
   ------------------------------------------------------------------------ */
int FROM=0;
int TO=10000;
int seed=0;
int EPOCH=200;

double allowable_error=0.0;
double error_limit = 0.0;
double learning_ratio = 0.1;
double rand_range = 0.1;
double sigmoid_slope = 1.7;

/*         THIS NETWORK IS THE THREE LAYERD PERCEPTRON */
/* 
   NOTE: 0-th neuron will be handled as a threshold value
   The actual number of neurons in each layer will be automatically
   settled by program in accordance with both the command line arguments
   and input file. 
*/
int N_of_input = 0;
int N_of_output = 0;

neuron *output, *input;

int NPATS = 0;               /* Number of patterns to be learned */
char *input_file_name = NULL;
char *teacher_file_name = NULL;
char *init_file_name = NULL;
matrix example, teacher;
matinfo Input_Info, Teacher_Info;

/* McCallogh and Pitts' formal neuron */ 
double formal_f(double state);

/* Logistic function $f(x)=1/(1+ exp(-x))$ */
double logistic_f(double state);
double Dlogistic_f(double state);

/* Hyperbolic function $f(x)= 1/2 * ( tanh(x) + 1) */
double hyperbolic_f(double state);
double Dhyperbolic_f(double state);

/* linear output function */
double linear_f(double state);
double Dlinear_f(double state);

/* initialize a neuron which has N inputs */
void initialize_neuron(neuron *a, int N, double (*output_func)(), double (*Doutput_func)(), double range);

/* copy whole contents from a to b */
void copy_neuron(neuron *a, neuron *b);

void calc_output(neuron *a);

void usage(char *prgname)
{
    fprintf(stderr, "### Program name: %s - A Backpropagation simulator\n", prgname);
    fprintf(stderr, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
    fprintf(stderr, "Usage: %s -input <filename> -teacher <filename> [-init <initial_weight_file_name>] [-error <double>] [-learn <double>] [-epoch <int>] [-to <int>] [-seed <int>] [-rand_range <double>] [-slope <double>] [-allowable_error <double>]\n", prgname);
    fprintf(stderr, "Arguments: \n");
    fprintf(stderr, "     input, input file name to be learned\n");
    fprintf(stderr, "     teacher, teacher file name to be learned\n");
    fprintf(stderr, "default settings of options: \n");
    fprintf(stderr, "         learn = %g: learning ratio\n", learning_ratio);
    fprintf(stderr, "         error = %g: error limit\n", error_limit);
    fprintf(stderr, "         allowable_error = %g: allowable error limit\n", allowable_error);
    fprintf(stderr, "         epoch = %d: epoch \n", EPOCH);
    fprintf(stderr, "         slope = %g: slope of sigmoid curve\n", sigmoid_slope);
    fprintf(stderr, "         to = %d: max iterations\n", TO);
    fprintf(stderr, "         rand_range: = %g: range for initial values\n", rand_range);
    fprintf(stderr, "         seed: seed of random number generator\n");
}

/* McCalloch and Pitts' formal neuron */ 
double formal_f(double state) 
{
    return (state > 0) ? 1.0 : 0.0;
}

/* Logistic function */
double logistic_f(double state)
{
    return 1.0 / ( 1.0 + exp( - sigmoid_slope * state ));
}

double Dlogistic_f(double state)
{
    return ((state) * (1.0 - state) * sigmoid_slope );
}

double linear_f(double state)
{
    return state;
}

double Dlinear_f(double state)
{
    return 1.0;
}

/* Initilazing a neuron, which is described as a struct NEURON */
void initialize_neuron(neuron *a, int N, double output_func(), double Doutput_func(), double range )
{
    int i;

    a->output = 0.0;
    a->error = 0.0;
    a->Ninp = N;
    a->output_func = output_func;
    a->Doutput_func = Doutput_func;

    if ( a->Ninp == 0 ) {
        a->inp_wgt = NULL;
        a->inp_neuron = NULL;
        return;
    }

    a->inp_wgt = (double *)malloc(sizeof(double) * (a->Ninp +1));
    if ( a->inp_wgt == NULL){
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i<= a->Ninp; i++) {
        a->inp_wgt[i] = 2.0 * MTgenrand_f() * range - range;
    }

    a->inp_neuron = (neuron **)malloc(sizeof(neuron *) * (a->Ninp +1));
    if ( a->inp_neuron == NULL ) {
        fprintf(stderr, "Can not allocate memory\n");
        exit (EXIT_FAILURE);
    }
    for (i=0; i <= a->Ninp; i++) {
        a->inp_neuron[i] = NULL;
    }
}

void print_neuron(neuron *a)
{
    int i;

    fprintf(stderr, "output :%5.3f  ", a->output);
    fprintf(stderr, "Ninp   :%d\n", a->Ninp);
    for (i=0; i< a->Ninp; i++) {
	fprintf(stderr, "\twgt[%d]=%5.3f \n", i, a->inp_wgt[i]);
    }
}

void copy_neuron(neuron *a, neuron *b)
{
    int i;

    b->output = a->output;
    b->Ninp = a->Ninp;
    b->error = a->error;
    
    for (i=0; i<= a->Ninp; i++) {
	b->inp_wgt[i] = a->inp_wgt[i];
        if ( a->inp_neuron[i] != NULL && b->inp_neuron[i] != NULL ){ 
            b->inp_neuron[i] = a->inp_neuron[i];
        }
    }
}

void calc_output (neuron *a) 
{
    double wrk = 0.0;
    int i;

    for(i=0; i<= a->Ninp; i++){
         wrk += a->inp_wgt[i] * a->inp_neuron[i]->output;
    }
    a->output = a->output_func(wrk);
}

int ParseArgs(int argc, char **argv)
/* Parse command line arguments and set the variables */
/* The variables to be set are
   sigmod_slope, randam_seed, error_limit,
   EPOCH, allowable_error, TO
   input_file_name, teacher_file_name, init_file_name */
{
    int argn=1;

    while ( argn < argc ) {
        if ((strcmp(argv[argn],"--help") == 0) || (strcmp(argv[argn],"-h") == 0)) {
            usage(argv[0]);     exit (EXIT_FAILURE);
        }
        if ((argv[argn][0] == '-') && (argv[argn+1] != NULL)) {
            if (strcmp(argv[argn], "-learn") == 0) {
                learning_ratio = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-rand_range") == 0) {
                rand_range = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-slope") == 0) {
                sigmoid_slope = atof(argv[++argn]);
            }
            if (strcmp(argv[argn],"-seed") == 0) {
                seed = atoi(argv[++argn]);
            }
            if (strcmp(argv[argn], "-error") == 0) {
                error_limit = atof(argv[++argn]);
            }
            if (strcmp(argv[argn], "-epoch") == 0) {
                EPOCH = atoi(argv[++argn]);
            }
	    if (strcmp(argv[argn], "-allowable_error") == 0) { 
		allowable_error = atof(argv[++argn]);
	    }
            if (strcmp(argv[argn], "-to") == 0) {
                TO = atoi(argv[++argn]);
		if ( TO < 0 ) {
		    usage(argv[0]);  
		    fprintf(stderr, "### !!!!! Invalid -to parameter !!!! \n");
		    exit (EXIT_FAILURE);
		}
            }
            if (strcmp(argv[argn], "-input") == 0) {
		input_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (input_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(input_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		input_file_name[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-teacher") == 0) {
		teacher_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (teacher_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(teacher_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		teacher_file_name[strlen(argv[argn])] = '\0';
	    }
            if (strcmp(argv[argn], "-init") == 0) {
		init_file_name = (char *)malloc(strlen(argv[++argn])+1);
		if (init_file_name == NULL) {
		    fprintf(stderr, "Could not allocate memory in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		if (strcpy(init_file_name, argv[argn]) == NULL) {
		    fprintf(stderr, "Could not copy filename in ParseArgs()\n");
		    exit (EXIT_FAILURE);
		}
		init_file_name[strlen(argv[argn])] = '\0';
	    }
	}
	argn++;
    }

    if (input_file_name == NULL) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid filename of input=%s !!!\n", input_file_name);
	exit (EXIT_FAILURE);
    }
    if (teacher_file_name == NULL) {
	usage(argv[0]);
	fprintf(stderr, "### !!! Invalid filename of teacher=%s !!!\n", teacher_file_name);
	exit (EXIT_FAILURE);
    }
    Input_Info = matinfo_read(input_file_name);
    Teacher_Info = matinfo_read(teacher_file_name);

    N_of_input = Input_Info->m;
    if ( N_of_input <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    N_of_output = Teacher_Info->m;
    if ( N_of_output <= 0 ) { usage(argv[0]); exit (EXIT_FAILURE);}

    NPATS = Input_Info->n;
    example = Input_Info->mat;
    teacher = Teacher_Info->mat;

    return OK;
}

/* Set values for an input pattern to  each input neron */
/* where pattern number will be given by second argument pat(int) */
void set_input(neuron *input, int pat) 
{
    int i;
    
    input[0].output = 1.0; /* for threshold unit */
    for (i=1; i<=N_of_input; i++) {
	input[i].output = *(example+offs(NPATS,N_of_input,pat+1,i));
    }
}

void calc_system_output(int pat)
{
    int i;

    set_input(input, pat);
    for (i=0; i< N_of_output; i++) {
	calc_output(&output[i]);
    }
}

double calc_SSE(void)
{
    int i,pat;
    double error=0.0, ret=0.0;

    for (pat=0; pat<NPATS; pat++) {
	calc_system_output(pat);
	for (i=0; i< N_of_output; i++) {
	    error = (output[i].output 
		     - *(teacher+offs(NPATS,N_of_output,pat+1,i+1)));
	    if (fabs(error) < allowable_error) { error = 0.0; }
	    output[i].error = error;
	    ret += error * error;
	}
    }
    return ret;
}

/* Print result to the stderr */
void print_result(neuron *output, neuron *input) 
{
    int pat, i,j;
    
    for (pat=0; pat<NPATS; pat++) {
	calc_system_output(pat);
	
        fprintf(stderr,"Pattern %d: (", pat);
        for (i=1; i<=N_of_input; i++) {
            fprintf(stderr," %+-5.2f", input[i].output);
        }
        fprintf(stderr,")->output(");
        for (i=0; i< N_of_output; i++) {
            fprintf(stderr,"%+-5.2f ", output[i].output);
        }
        fprintf(stderr,")=teacher(");
        for (i=0; i< N_of_output; i++) {
            fprintf(stderr,"%+-5.2f ", *(teacher+offs(NPATS,N_of_output,pat+1,i+1)));
        }
        fprintf(stderr,")\n");
	
        for (i=0; i< N_of_output; i++) {
            fprintf(stderr,"  output[%d] f(", i);
            for (j=0; j<=N_of_input; j++) {
                fprintf(stderr," %+-5.2f * %+-5.2f", output[i].inp_wgt[j], 
		                             output[i].inp_neuron[j]->output);
            }
            fprintf(stderr,") =%5.3f\n", output[i].output);
        }
    }
}

/* Print weight matrix to the stdout */
void print_weight(neuron *output, neuron *input) 
{
    int i,j;
	
    for (i=0; i< N_of_output; i++) {
	fprintf(stdout,"###  output[%d] \n", i);
	for (j=0; j<=N_of_input; j++) {
	    fprintf(stdout," %+-9.6f ", output[i].inp_wgt[j]);
	}
	fprintf(stdout,"\n");
    }
}

void Initialize_network(void) 
{
    int i, j;
    FILE *fp;
    matrix TMP;

    if (error_limit == 0.0 ) {
	error_limit = 0.05;
    }

    input = (neuron *)malloc(sizeof(neuron) * (N_of_input+1));
    if ( input == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }
    output = (neuron *)malloc(sizeof(neuron) * (N_of_output));
    if ( output == NULL ) {
	fprintf(stderr, "Could not allocate memroy in Initialize_network()\n");
	exit (EXIT_FAILURE);
    }
    
    /* initialize an output neuron */
    for (i=0; i< N_of_output; i++) {
        initialize_neuron(&output[i],N_of_input,logistic_f,Dlogistic_f, rand_range);
    }
    
    /* initialize each neuron in input layer */
    /* input[0] will be treated as a threshold */
    initialize_neuron(&input[0],0, logistic_f, Dlogistic_f, rand_range);   
    input[0].output = 1.0;
    for (i=1; i<=N_of_input; i++) { 
        initialize_neuron(&input[i],0, logistic_f, Dlogistic_f, rand_range); 
    }
    /* NOTE: */    
    /* The output values input[0] is dealed as threshold value,     */
    /* therefore, the output must be constantly set to 1.           */
    
    /********* Making connection among neurons **********/
    for (i=0; i< N_of_output; i++) {
	for (j=0; j<=N_of_input; j++) {
	    output[i].inp_neuron[j] = &input[j]; 
	}
    }
    
    /* When init_file_name will be given, then          */
    /* read the initial value of weight value from file */
    if ( init_file_name != NULL ) {
	TMP = matinit(1,N_of_input+1);
	fp = fileio_util_openr(init_file_name);

	TMP = matinit(1,N_of_input+1);
        for (i=0; i< N_of_output; i++) {
	    freadmat(fp, TMP, 1, N_of_input+1);
	    for (j=0; j<=N_of_input; j++) {
                output[i].inp_wgt[j] = *(TMP+offs(1,N_of_input+1,1,j+1));
	    }
        }
	fileio_util_close(fp);
    }
}

/* Compute weight values */
void update_weight(neuron *a)
{
    int i;
    double w;

    for (i=0; i<=a->Ninp; i++) {
	w = - learning_ratio * a->error * a->Doutput_func(a->output) * a->inp_neuron[i]->output;
	a->inp_wgt[i] +=  w;
    }
}

void do_shuffle(int *order,int n)
{
    int i, index, tmp;

    for (i=0; i<n; i++) {
        index = i + (int)(MTgenrand_f() * ((double)n - (double)i ));
	tmp = order[index];
	order[index] = order[i];
	order[i] = tmp;
    }
}

/* main loop of computation */
int main_loop (double *SSE) 
{
    int t,i, pat;
    double error;
    int *shuffle;
    double mse;

    shuffle=(int *)malloc(sizeof(int) * NPATS);
    if (shuffle == NULL) {
	fprintf(stderr, "Could not allocate memory for shuffle\n");
	exit (EXIT_FAILURE);
    }
    for ( i=0; i<NPATS; i++ ) {
	shuffle[i] = i;
    }
    
    for (t=FROM; t<TO; t++){ /* learning time */

	*SSE = calc_SSE();
	mse = *SSE / ((double)NPATS * (double)N_of_output);
	mse = sqrt(mse);
	if ( mse <= error_limit ){
		return t;
	}

        if ( (t % EPOCH) == 0 ){
	    fprintf(stderr,"%05d %g ### MSE\n", t, mse);
        }

	do_shuffle(shuffle,NPATS);

        for (pat=0; pat<NPATS; pat++){ /* foreach pattern */
	    calc_system_output(shuffle[pat]);
            for (i=0; i< N_of_output; i++) {
                error = (output[i].output 
			 - *(teacher+offs(NPATS,N_of_output,shuffle[pat]+1,i+1)));
		if (fabs(error) < allowable_error) { error = 0.0; }
		output[i].error = error;
		update_weight(&output[i]); 
            }
         }
    }
    return t;
}

int main(int argc, char **argv) 
{
    long int iterations = 0;
    double SSE=0.0;
    double mse;

    ParseArgs(argc, argv);

    init_random(seed);
    Initialize_network();

    iterations = main_loop(&SSE);
    print_result(output, input);
    print_weight(output, input);

    fprintf(stdout, "### %s -input %s -teacher %s %s %s -learn %g -error %g -slope %g -epoch %d -to %d -seed %d -rand_range %g -allowable_error %g\n",
	    argv[0], input_file_name, teacher_file_name, 
            (init_file_name != NULL) ? "-init" : "",
            (init_file_name != NULL) ? init_file_name : "",
	    learning_ratio, error_limit, sigmoid_slope, EPOCH, TO, 
	    get_seed(), rand_range, allowable_error);

    if ( iterations == TO ) {
	fprintf(stdout,"### COULD NOT LEARN AT ALL WITHIN %ld TIMES\n", 
		iterations);
    } else {
	fprintf(stdout,"### Learning completed in %ld time(s)\n", 
		iterations);
    }

    SSE = calc_SSE();
    mse = SSE / ((double)NPATS * (double)N_of_output);
    mse = sqrt(mse);
    fprintf(stdout,"### Final MSE = %g at iter=%ld\n", mse, iterations);
    return OK;
}
