/* A C-program for MT19937: Real number version                */
/*   genrand() generates one pseudorandom real number (double) */
/* which is uniformly distributed on [0,1]-interval, for each  */
/* call. sgenrand(seed) set initial values to the working area */
/* of 624 words. Before genrand(), sgenrand(seed) must be      */
/* called once. (seed is any 32-bit integer except for 0).     */
/* Integer generator is obtained by modifying two lines.       */
/*   Coded by Takuji Nishimura, considering the suggestions by */
/* Topher Cooper and Marc Rieffel in July-Aug. 1997.           */

/* This library is free software; you can redistribute it and/or   */
/* modify it under the terms of the GNU Library General Public     */
/* License as published by the Free Software Foundation; either    */
/* version 2 of the License, or (at your option) any later         */
/* version.                                                        */
/* This library is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            */
/* See the GNU Library General Public License for more details.    */
/* You should have received a copy of the GNU Library General      */
/* Public License along with this library; if not, write to the    */
/* Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA   */ 
/* 02111-1307  USA                                                 */

/* Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.       */
/* When you use this, send an email to: matumoto@math.keio.ac.jp   */
/* with an appropriate reference to your work.                     */

#include <stdio.h>
#include "general.h"
#include "rand_gen.h"
#include <time.h>

#ifdef UNIFORM_RAND
 #include <stdlib.h>
 #define STANDALONE
#endif
 #ifdef INT_UNIFORM_RAND
 #include <stdlib.h>
 #define STANDALONE
#endif
#ifdef UNIFORM_RAND2
 #include <stdlib.h>
 #define STANDALONE
#endif

long loadingtime;
unsigned int SEED;
FLAG IsSeedInitialized = False;

/* Period parameters */  
#define N 624
#define M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* Tempering parameters */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

/* initializing the array with a NONZERO seed */
void MTsgenrand (unsigned long seed) 
{

    if (IsSeedInitialized) return;

    /* setting initial seeds to mt[N] using         */
    /* the generator Line 25 of Table 1 in          */
    /* [KNUTH 1981, The Art of Computer Programming */
    /*    Vol. 2 (2nd Ed.), pp102]                  */
    mt[0]= seed & 0xffffffff;
    for (mti=1; mti<N; mti++)
        mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;

    SEED = seed;
    IsSeedInitialized = True;
}

/* initialize of random number sequence */
int init_random(unsigned long arg) 
{
    if ( arg == 0 ) {
        SEED = (unsigned)time(&loadingtime); 
    } else {
        SEED = arg;
    }
    MTsgenrand(SEED);
    return OK;
}

unsigned int get_seed(void){
    return SEED;
}

double /* generating reals */
/* unsigned long */ /* for integer generation */
MTgenrand_f (void) 
{
    unsigned long y;
    static unsigned long mag01[2]={0x0, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if ( ! IsSeedInitialized ) {
        SEED = (unsigned)time(&loadingtime);
        MTsgenrand(SEED);
		}

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if sgenrand() has not been called, */
            MTsgenrand(4357); /* a default initial seed is used   */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

        mti = 0;
    }
  
    y = mt[mti++];
    y ^= TEMPERING_SHIFT_U(y);
    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
    y ^= TEMPERING_SHIFT_L(y);

    return ( (double)y * 2.3283064370807974e-10 ); /* reals */
    /* return y; */ /* for integer generation */
}

double /* generating reals */
/* unsigned long */ /* for integer generation */
MTgenrand_fh (void) {
    unsigned long y;
    static unsigned long mag01[2]={0x0, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if sgenrand() has not been called, */
            MTsgenrand(4357); /* a default initial seed is used   */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

        mti = 0;
    }
  
    y = mt[mti++];
    y ^= TEMPERING_SHIFT_U(y);
    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
    y ^= TEMPERING_SHIFT_L(y);

    return ( (double)y * 2.3283064365386963e-10 ); /* reals: [0,1)-interval */
    /* return y; */ /* for integer generation */
}

unsigned long MTgenrand_i (void) {
    unsigned long y;
    static unsigned long mag01[2]={0x0, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if sgenrand() has not been called, */
            MTsgenrand(4357); /* a default initial seed is used   */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

        mti = 0;
    }
  
    y = mt[mti++];
    y ^= TEMPERING_SHIFT_U(y);
    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
    y ^= TEMPERING_SHIFT_L(y);

    return y; 
}

#ifdef STANDALONE

int usage(void){
#ifdef UNIFORM_RAND
    fprintf(stderr, "randgen(MTrand.c) : random number geneator\n");
    fprintf(stderr, "Usage: uniform_rand N M [SEED]\n");
#endif
#ifdef INT_UNIFORM_RAND
    fprintf(stderr, "int_uniform_rand(MTrand.c) : random number geneator\n");
    fprintf(stderr, "Usage: int_uniform_rand N M MIN MAX [SEED]\n");
#endif
#ifdef UNIFORM_RAND2
    fprintf(stderr, "randgen2(MTrand.c) : random number geneator\n");
    fprintf(stderr, "Usage: uniform_rand2 N M [SEED]\n");
#endif
    exit (EXIT_FAILURE);
}

int main (int argc, char **argv)
{
    int i,j;
    int n,m;

#if INT_UNIFORM_RAND
    int MIN,MAX;
    if ( argc < 5 ) usage();
#else
    if ( argc < 3 ) usage();
#endif

    n = atoi(argv[1]);
    if (n <= 0) n = 10;
 
    m = atoi(argv[2]);
    if (m <= 0) m = 10;

#ifdef INT_UNIFORM_RAND
    MIN = atoi(argv[3]);
    if (MIN < 0) MIN = 0;
    MAX = atoi(argv[4]);
    if (MAX < 0) MAX = MIN+1;
#endif INT_UNIFORM_RAND

#ifdef INT_UNIFORM_RAND
    if (argc == 6) SEED = atoi(argv[5]);
#else
    if (argc == 4) SEED = atoi(argv[3]);
#endif irandge
    else SEED = (unsigned)time(&loadingtime);
    MTsgenrand(SEED);
    /*    MTsgenrand(4357); */ /* any nonzero integer can be used as a seed */

#ifdef UNIFORM_RAND
	fprintf(stdout,"### ProgramName: uniform_rand(C version) %d %d %d\n",n,m,SEED);
#endif UNIFORM_RAND

#ifdef INT_UNIFORM_RAND
	fprintf(stdout,"### ProgramName: int_uniform_rand(C version) %d %d %d %d %d\n",n,m,MIN,MAX,SEED);
#endif INT_UNIFORM_RAND

#ifdef UNIFORM_RAND2
	fprintf(stdout,"### ProgramName: uniform_rand2(C version) %d %d %d\n",n,m,SEED);
#endif randgein2
	fprintf(stdout, "### Creator: Shinichi ASAKAWA <asakawa@twcu.ac.jp>\n");
	(void)time(&loadingtime);
	fprintf(stdout, "### CreationDate: %s", ctime(&loadingtime));

    for (i=0; i<n; i++) {
        for (j=0; j<m; j++) {
#ifdef UNIFORM_RAND
            printf("%5f ", MTgenrand_f());
#endif UNIFORM_RAND

#ifdef UNIFORM_RAND2
            printf("%5f ", MTgenrand_fh());
#endif
#ifdef INT_UNIFORM_RAND
/*
	  printf("%12u ", MTgenrand_i() % (MAX - MIN) + MIN);
*/
            printf("%d ",((unsigned int)MTgenrand_i() % (MAX + 1 - MIN)) + MIN);
#endif INT_UNIFORM_RAND
        }
        printf("\n");
    }
    return OK;
}
#endif STANDALONE
